#ifndef WRP_CAE_CORE_AUTOGEN_METHODS_H_
#define WRP_CAE_CORE_AUTOGEN_METHODS_H_

#include <chimaera/chimaera.h>

/**
 * Auto-generated method definitions for core
 */

namespace wrp_cae::core {

namespace Method {
// Inherited methods
GLOBAL_CONST chi::u32 kCreate = 0;
GLOBAL_CONST chi::u32 kDestroy = 1;

// core-specific methods
GLOBAL_CONST chi::u32 kParseOmni = 10;
}  // namespace Method

}  // namespace wrp_cae::core

#endif  // WRP_CAE_CORE_AUTOGEN_METHODS_H_
