spack load \
catch2@3.0.1 \
yaml-cpp \
doxygen \
libfabric \
libelf \
mochi-thallium+cereal@0.10.1 \
argobots@1.1+affinity \
cereal \
boost@1.7: +context +fiber +coroutine +regex +system +filesystem +serialization +pic +math \
mpi \
py-pybind11 \
lzo \
bzip2 \
zstd \
lz4 \
zlib \
xz \
brotli \
snappy \
c-blosc2
