// Copyright 2025 Chimaera Project
// Licensed under the Apache License, Version 2.0

#include "chimaera/integer_timer.h"

namespace chi {

// Initialize static member
u64 IntegerTimer::current_time_us_ = 0;

}  // namespace chi
